Unit SWCZip; {Erstdatum: DownLoad vom 29.04.1998}


{In fact it requires that you pass the "LZHPack" Procedure 2 procedural
 parameter of Type 'GetProcType' and 'PutProcType' (declared below)
 which will accept 3 parameters and act in every way like a
 'BlockRead'/'BlockWrite' Procedure call. Your 'GetProcType' Procedure
 should return the data to be Compressed, and Your 'PutProcType' Procedure
 should do something with the Compressed data (ie., put it in a File).  In
 Case you need to know (and you do if you want to deCompress this data
 again) the number of Bytes in the Compressed data (original, not
 Compressed size) is returned in 'Bytes_Written'.

 GetBytesProc = Procedure(Var DTA; NBytes:Word; Var Bytes_Got : Word);

 DTA is the start of a memory location where the inFormation returned should
 be.  NBytes is the number of Bytes requested.  The actual number of Bytes
 returned must be passed in Bytes_Got (if there is no more data then 0
 should be returned).

 PutBytesProc = Procedure(Var DTA; NBytes:Word; Var Bytes_Got : Word);

 As above except instead of asking For data the Procedure is dumping out
 Compressed data, do somthing With it.

 "LZHUnPack" is basically the same thing in reverse.  It requires
 procedural parameters of Type 'PutProcType'/'GetProcType' which
 will act as above.  'GetProcType' must retrieve data Compressed using
 "LZHPack" (above) and feed it to the unpacking routine as requested.
 'PutProcType' must accept the deCompressed data and do something
 withit.  You must also pass in the original size of the deCompressed data,
 failure to do so will have adverse results.


  Don't Forget that as procedural parameters the 'GetProcType'/'PutProcType'
  Procedures must be Compiled in the 'F+' state to avoid a catastrophe.
}

{ note: All the large data structures For these routines are allocated when
  needed from the heap, and deallocated when finished.  So when not in use
  memory requirements are minimal.  However, this Unit Uses about 34K of
  heap space, and 400 Bytes of stack when in use. }


Interface


Uses SWCPGX,
     SWCSubPr,
     SWCVaria;


Const
  ZIPUNZIPC : IntL=-001;

Type
  GetBytesProc = Procedure(Pos,PEnd: IntL;Var DTA: Byte;NBytes: Word;Var Bytes_Got: Word);
  PutBytesProc = Procedure(Pos,PEnd: IntL;Var DTA: Byte;NBytes: Word;Var Bytes_Put: Word);



Procedure BlockGet(Pos,PEnd: IntL;Var Target: Byte;CountBytes: Word;Var Actual_Bytes: Word);
Procedure BlockPut(Pos,PEnd: IntL;Var Source: Byte;CountBytes: Word;Var Actual_Bytes: Word);
Function  IsFileZipped(ZPath: ST79): Bool;
Procedure LZHPack(Var Bytes_Written : LongInt;
          GetBytes : GetBytesProc;PutBytes : PutBytesProc);
Procedure LZHUnpack(TextSize : LongInt;
          GetBytes : GetBytesProc;PutBytes : PutBytesProc);
Function  UnZip(SPath,TPath: ST79): Bool;
Function  Zip(SPath,TPath: ST79;BAsk,BMZip: Bool): Bool;



Implementation



Const
  PosNG : Word=000;
  PosNP : Word=000;
  Buf   : Word=000;
  MaxBuf    = 32768; {Must be bigger than the biggest chunk being asked For.}

Var
  BVersion3 : bool;
  BWritten  : IntL;
  HIn       : Hand;
  HOut      : Hand;
  S         : STFF;
  SegIn     : Word;
  SegOut    : Word;
  SegIO     : Word;
  SHWB0     : Bool;
  SHWH0     : Hand;
  Size      : IntL;
  Temp      : Word;
  ZActivity : ST13;



Const
  Exit_OK = 0;
  Exit_FAILED = 1;

  { LZSS Parameters }
  N = 4096;                            { Size of String buffer }
  F = 60;                              { Size of look-ahead buffer }
  THRESHOLD = 2;
  NUL = N;                             { end of tree's node  }

  { Huffman coding parameters }
  N_Char = (256 - THRESHOLD + F);

  { Character code (:= 0..N_Char-1) }
  T = (N_Char * 2 - 1);                { Size of table }
  R = (T - 1);                         { root position }

  { update when cumulative frequency }
  { reaches to this value }
  MAX_FREQ = $8000;

{
 * Tables For encoding/decoding upper 6 bits of
 * sliding dictionary Pointer
 }

  { encoder table }
  p_len : Array[0..63] of Byte =
  ($03, $04, $04, $04, $05, $05, $05, $05,
   $05, $05, $05, $05, $06, $06, $06, $06,
   $06, $06, $06, $06, $06, $06, $06, $06,
   $07, $07, $07, $07, $07, $07, $07, $07,
   $07, $07, $07, $07, $07, $07, $07, $07,
   $07, $07, $07, $07, $07, $07, $07, $07,
   $08, $08, $08, $08, $08, $08, $08, $08,
   $08, $08, $08, $08, $08, $08, $08, $08);

  p_code : Array[0..63] of Byte =
  ($00, $20, $30, $40, $50, $58, $60, $68,
   $70, $78, $80, $88, $90, $94, $98, $9C,
   $A0, $A4, $A8, $AC, $B0, $B4, $B8, $BC,
   $C0, $C2, $C4, $C6, $C8, $CA, $CC, $CE,
   $D0, $D2, $D4, $D6, $D8, $DA, $DC, $DE,
   $E0, $E2, $E4, $E6, $E8, $EA, $EC, $EE,
   $F0, $F1, $F2, $F3, $F4, $F5, $F6, $F7,
   $F8, $F9, $FA, $FB, $FC, $FD, $FE, $FF);

  { decoder table }
  d_code : Array[0..255] of Byte =
  ($00, $00, $00, $00, $00, $00, $00, $00,
   $00, $00, $00, $00, $00, $00, $00, $00,
   $00, $00, $00, $00, $00, $00, $00, $00,
   $00, $00, $00, $00, $00, $00, $00, $00,
   $01, $01, $01, $01, $01, $01, $01, $01,
   $01, $01, $01, $01, $01, $01, $01, $01,
   $02, $02, $02, $02, $02, $02, $02, $02,
   $02, $02, $02, $02, $02, $02, $02, $02,
   $03, $03, $03, $03, $03, $03, $03, $03,
   $03, $03, $03, $03, $03, $03, $03, $03,
   $04, $04, $04, $04, $04, $04, $04, $04,
   $05, $05, $05, $05, $05, $05, $05, $05,
   $06, $06, $06, $06, $06, $06, $06, $06,
   $07, $07, $07, $07, $07, $07, $07, $07,
   $08, $08, $08, $08, $08, $08, $08, $08,
   $09, $09, $09, $09, $09, $09, $09, $09,
   $0A, $0A, $0A, $0A, $0A, $0A, $0A, $0A,
   $0B, $0B, $0B, $0B, $0B, $0B, $0B, $0B,
   $0C, $0C, $0C, $0C, $0D, $0D, $0D, $0D,
   $0E, $0E, $0E, $0E, $0F, $0F, $0F, $0F,
   $10, $10, $10, $10, $11, $11, $11, $11,
   $12, $12, $12, $12, $13, $13, $13, $13,
   $14, $14, $14, $14, $15, $15, $15, $15,
   $16, $16, $16, $16, $17, $17, $17, $17,
   $18, $18, $19, $19, $1A, $1A, $1B, $1B,
   $1C, $1C, $1D, $1D, $1E, $1E, $1F, $1F,
   $20, $20, $21, $21, $22, $22, $23, $23,
   $24, $24, $25, $25, $26, $26, $27, $27,
   $28, $28, $29, $29, $2A, $2A, $2B, $2B,
   $2C, $2C, $2D, $2D, $2E, $2E, $2F, $2F,
   $30, $31, $32, $33, $34, $35, $36, $37,
   $38, $39, $3A, $3B, $3C, $3D, $3E, $3F);

  d_len : Array[0..255] of Byte =
  ($03, $03, $03, $03, $03, $03, $03, $03,
   $03, $03, $03, $03, $03, $03, $03, $03,
   $03, $03, $03, $03, $03, $03, $03, $03,
   $03, $03, $03, $03, $03, $03, $03, $03,
   $04, $04, $04, $04, $04, $04, $04, $04,
   $04, $04, $04, $04, $04, $04, $04, $04,
   $04, $04, $04, $04, $04, $04, $04, $04,
   $04, $04, $04, $04, $04, $04, $04, $04,
   $04, $04, $04, $04, $04, $04, $04, $04,
   $04, $04, $04, $04, $04, $04, $04, $04,
   $05, $05, $05, $05, $05, $05, $05, $05,
   $05, $05, $05, $05, $05, $05, $05, $05,
   $05, $05, $05, $05, $05, $05, $05, $05,
   $05, $05, $05, $05, $05, $05, $05, $05,
   $05, $05, $05, $05, $05, $05, $05, $05,
   $05, $05, $05, $05, $05, $05, $05, $05,
   $05, $05, $05, $05, $05, $05, $05, $05,
   $05, $05, $05, $05, $05, $05, $05, $05,
   $06, $06, $06, $06, $06, $06, $06, $06,
   $06, $06, $06, $06, $06, $06, $06, $06,
   $06, $06, $06, $06, $06, $06, $06, $06,
   $06, $06, $06, $06, $06, $06, $06, $06,
   $06, $06, $06, $06, $06, $06, $06, $06,
   $06, $06, $06, $06, $06, $06, $06, $06,
   $07, $07, $07, $07, $07, $07, $07, $07,
   $07, $07, $07, $07, $07, $07, $07, $07,
   $07, $07, $07, $07, $07, $07, $07, $07,
   $07, $07, $07, $07, $07, $07, $07, $07,
   $07, $07, $07, $07, $07, $07, $07, $07,
   $07, $07, $07, $07, $07, $07, $07, $07,
   $08, $08, $08, $08, $08, $08, $08, $08,
   $08, $08, $08, $08, $08, $08, $08, $08);

  getbuf : Word = 0;
  getlen : Byte = 0;
  putlen : Byte = 0;
  putbuf : Word = 0;
  TextSize : LongInt = 0;
  codesize : LongInt = 0;
  printcount : LongInt = 0;
  match_position : Integer = 0;
  match_length : Integer = 0;


Type
  FreqType = Array[0..T] of Word;
  FreqPtr = ^FreqType;
  PntrType = Array[0..pred(T + N_Char)] of Integer;
  pntrPtr = ^PntrType;
  SonType = Array[0..pred(T)] of Integer;
  SonPtr = ^SonType;
  {TextBufType = Array[0..N + F - 2] of Byte;}
  TextBufType = Array[0..N + F - 1] of Byte;
  TBufPtr = ^TextBufType;
  WordRay = Array[0..N] of Integer;
  WordRayPtr = ^WordRay;
  BWordRay = Array[0..N + 256] of Integer;
  BWordRayPtr = ^BWordRay;

Var
  Text_buf : TBufPtr;
  lson, dad : WordRayPtr;
  rson : BWordRayPtr;
  freq : FreqPtr;                      { cumulative freq table }

{
 * pointing parent nodes.
 * area [T..(T + N_Char - 1)] are Pointers For leaves
 }
  prnt : pntrPtr;

  { pointing children nodes (son[], son[] + 1)}
  son : SonPtr;


Procedure BlockGet(Pos,PEnd: IntL;Var Target: Byte;CountBytes: Word;Var Actual_Bytes: Word);
Var
  Temp  : Word;

Procedure HelpScramble;
Var
  Byte0 : Byte;
Begin
  If ZActivity='Kompression' Then
  Begin
    If Not BVersion3 Then
    Begin
      For Byte0:=001 To Length(Scramble) Do
      Target:=CrypTab[Ord(Scramble[Byte0])]+Target;
    End
    Else
    If ((FilePos(HIn)<=MaxBuf) And (PosNG>ZIPUNZIPC))
    Or (FilePos(HIn)>MaxBuf) Then
    For Byte0:=001 To Length(Scramble) Do
    Target:=CrypTab[Ord(Scramble[Byte0])]+Target;
  End;
End;

Begin

  If CountBytes>000 Then
  Begin
    If (PosNG=Buf) Or (PosNG+CountBytes>Buf) Then
    Begin
      If PosNG=Buf Then
      Begin
        BlockRead(HIn,Mem[SegIn:0],MaxBuf,Buf);
        If ZActivity='Kompression' Then
        ShowP(ZActivity,SHWB0,SHWH0,020,000,000,
        FilePos(HIn),FileSize(HIn),BF);
      End
      Else
      Begin
        Move(Mem[SegIn:PosNG],Mem[SegIn:0],(Buf-PosNG)-001);
        BlockRead(HIn,Mem[SegIn:(Buf-PosNG)-001],
        MaxBuf-((Buf-PosNG)-001),Temp);
        Buf:=(Buf-PosNG)-001+Temp;
        If ZActivity='Kompression' Then
        ShowP(ZActivity,SHWB0,SHWH0,020,000,000,
        FilePos(HIn),FileSize(HIn),BF);
      End;
      PosNG:=000;
      If Buf=000 Then
      Begin
        Actual_Bytes:=000;
        Exit;
      End;
    End;
    Target:=Mem[SegIn:PosNG];
    HelpScramble;
    Inc(PosNG,CountBytes);
  End;
  If PosNG>Buf Then
  Actual_Bytes:=CountBytes-(PosNG-Buf)
  Else
  Actual_Bytes:=CountBytes;

End;




Procedure BlockPut(Pos,PEnd: IntL;Var Source: Byte;CountBytes: Word;Var Actual_Bytes: Word);
Var
  Temp  : Word;

Procedure HelpUnScramble;
Var
  Byte0 : Byte;
Begin
  If (ZActivity='Dekompression') And (Scramble<>'') Then
  Begin
    If Not BVersion3 Then
    Begin
      For Byte0:=Length(Scramble) DownTo 001 Do
      Source:=Source-CrypTab[Ord(Scramble[Byte0])];
    End
    Else
    If Pos>ZIPUNZIPC Then
    For Byte0:=Length(Scramble) DownTo 001 Do
    Source:=Source-CrypTab[Ord(Scramble[Byte0])];
  End;
End;

Begin

  If CountBytes>000 Then                { Flush condition }
  Begin
    If (PosNP=MaxBuf) Or (PosNP+CountBytes>MaxBuf) Then
    Begin
      If ZActivity='Dekompression' Then
      ShowP(ZActivity,SHWB0,SHWH0,020,000,000,Pos,PEnd,BF);
      BlockWrite(HOut,Mem[SegOut:0],PosNP,Temp);
      PosNP:=000;
    End;
    HelpUnScramble;
    Mem[SegOut:PosNP]:=Source;
    Inc(PosNP,CountBytes);
  End
  Else
  Begin
    BlockWrite(HOut,Mem[SegOut:0],PosNP,Temp);
    PosNP:=000;
  End;
  Actual_Bytes:=CountBytes;

End;




  Procedure InitTree;                  { Initializing tree }
  Var
    i : Integer;
  begin
    For i := N + 1 to N + 256 do
      rson^[i] := NUL;                 { root }
    For i := 0 to N do
      dad^[i] := NUL;                  { node }
  end;


  Procedure InsertNode(R : Integer);   { Inserting node to the tree }
  Var
    tmp, i, p, cmp : Integer;
    key : TBufPtr;
    c : Word;
  begin
    cmp := 1;
    key := @Text_buf^[R];
    p := succ(N) + key^[0];
    rson^[R] := NUL;
    lson^[R] := NUL;
    match_length := 0;
    While match_length < F do
      begin
        if (cmp >= 0) then
          begin
            if (rson^[p] <> NUL) then
              p := rson^[p]
            else
              begin
                rson^[p] := R;
                dad^[R] := p;
                Exit;
              end;
          end
        else
          begin
            if (lson^[p] <> NUL) then
              p := lson^[p]
            else
              begin
                lson^[p] := R;
                dad^[R] := p;
                Exit;
              end;
          end;
        i := 0;
        cmp := 0;
        While (i < F) and (cmp = 0) do
          begin
            inc(i);
            cmp := key^[i] - Text_buf^[p + i];
          end;
        if (i > THRESHOLD) then
          begin
            tmp := pred((R - p) and pred(N));
            if (i > match_length) then
              begin
                match_position := tmp;
                match_length := i;
              end;
            if (match_length < F) and (i = match_length) then
              begin
                c := tmp;
                if (c < match_position) then
                  match_position := c;
              end;
          end;
      end;                             { While True do }
    dad^[R] := dad^[p];
    lson^[R] := lson^[p];
    rson^[R] := rson^[p];
    dad^[lson^[p]] := R;
    dad^[rson^[p]] := R;
    if (rson^[dad^[p]] = p) then
      rson^[dad^[p]] := R
    else
      lson^[dad^[p]] := R;
    dad^[p] := NUL;                    { remove p }
  end;


  Procedure DeleteNode(p : Integer);   { Deleting node from the tree }
  Var
    q : Integer;
  begin
    if (dad^[p] = NUL) then
      Exit;                            { unregistered }
    if (rson^[p] = NUL) then
      q := lson^[p]
    else if (lson^[p] = NUL) then
      q := rson^[p]
    else
      begin
        q := lson^[p];
        if (rson^[q] <> NUL) then
          begin
            Repeat
              q := rson^[q];
            Until (rson^[q] = NUL);
            rson^[dad^[q]] := lson^[q];
            dad^[lson^[q]] := dad^[q];
            lson^[q] := lson^[p];
            dad^[lson^[p]] := q;
          end;
        rson^[q] := rson^[p];
        dad^[rson^[p]] := q;
      end;
    dad^[q] := dad^[p];
    if (rson^[dad^[p]] = p) then
      rson^[dad^[p]] := q
    else
      lson^[dad^[p]] := q;
    dad^[p] := NUL;
  end;

  { Huffman coding parameters }

  Function GetBit(GetBytes : GetBytesProc) : Integer; { get one bit }
  Var
    i : Byte;
    i2 : Integer;
    result : Word;
  begin
    While (getlen <= 8) do
      begin
        GetBytes(0,0,i, 1, result);
        if result = 1 then
          i2 := i
        else i2 := 0;
        getbuf := getbuf or (i2 shl (8 - getlen));
        inc(getlen, 8);
      end;
    i2 := getbuf;
    getbuf := getbuf shl 1;
    dec(getlen);
    GetBit := Integer((i2 < 0));
  end;


  Function GetByte(GetBytes : GetBytesProc) : Integer; { get a Byte }
  Var
    j : Byte;
    i, result : Word;
  begin
    While (getlen <= 8) do
      begin
        GetBytes(0,0,j, 1, result);
        if result = 1 then
          i := j
        else
          i := 0;
        getbuf := getbuf or (i shl (8 - getlen));
        inc(getlen, 8);
      end;
    i := getbuf;
    getbuf := getbuf shl 8;
    dec(getlen, 8);
    GetByte := Integer(i shr 8);
  end;


  Procedure Putcode(l : Integer; c : Word;
                    PutBytes : PutBytesProc); { output c bits }
  Var
    Temp : Byte;
    Got : Word;
  begin
    putbuf := putbuf or (c shr putlen);
    inc(putlen, l);
    if (putlen >= 8) then
      begin
        Temp := putbuf shr 8;
        PutBytes(0,0,Temp, 1, Got);
        dec(putlen, 8);
        if (putlen >= 8) then
          begin
            Temp := lo(putbuf);
            PutBytes(0,0,Temp, 1, Got);
            inc(codesize, 2);
            dec(putlen, 8);
            putbuf := c shl (l - putlen);
          end
        else
          begin
            putbuf := putbuf shl 8;
            inc(codesize);
          end;
      end;
  end;


  { initialize freq tree }

  Procedure StartHuff;
  Var
    i, j : Integer;
  begin
    For i := 0 to pred(N_Char) do
      begin
        freq^[i] := 1;
        son^[i] := i + T;
        prnt^[i + T] := i;
      end;
    i := 0;
    j := N_Char;
    While (j <= R) do
      begin
        freq^[j] := freq^[i] + freq^[i + 1];
        son^[j] := i;
        prnt^[i] := j;
        prnt^[i + 1] := j;
        inc(i, 2);
        inc(j);
      end;
    freq^[T] := $ffff;
    prnt^[R] := 0;
  end;


  { reConstruct freq tree }

  Procedure reConst;
  Var
    i, j, k, tmp : Integer;
    F, l : Word;
  begin
    { halven cumulative freq For leaf nodes }
    j := 0;
    For i := 0 to pred(T) do
      begin
        if (son^[i] >= T) then
          begin
            freq^[j] := succ(freq^[i]) div 2; {@@ Bug Fix MOD -> div @@}
            son^[j] := son^[i];
            inc(j);
          end;
      end;
    { make a tree : first, connect children nodes }
    i := 0;
    j := N_Char;
    While (j < T) do
      begin
        k := succ(i);
        F := freq^[i] + freq^[k];
        freq^[j] := F;
        k := pred(j);
        While F < freq^[k] do
          dec(k);
        inc(k);
        l := (j - k) shl 1;
        tmp := succ(k);
        move(freq^[k], freq^[tmp], l);
        freq^[k] := F;
        move(son^[k], son^[tmp], l);
        son^[k] := i;
        inc(i, 2);
        inc(j);
      end;
    { connect parent nodes }
    For i := 0 to pred(T) do
      begin
        k := son^[i];
        if (k >= T) then
          begin
            prnt^[k] := i;
          end
        else
          begin
            prnt^[k] := i;
            prnt^[succ(k)] := i;
          end;
      end;
  end;


  { update freq tree }

  Procedure update(c : Integer);
  Var
    i, j, k, l : Integer;
  begin
    if (freq^[R] = MAX_FREQ) then
      begin
        reConst;
      end;
    c := prnt^[c + T];
    Repeat
      inc(freq^[c]);
      k := freq^[c];
      { swap nodes to keep the tree freq-ordered }
      l := succ(c);
      if (k > freq^[l]) then
        begin
          While (k > freq^[l]) do
            inc(l);
          dec(l);
          freq^[c] := freq^[l];
          freq^[l] := k;
          i := son^[c];
          prnt^[i] := l;
          if (i < T) then prnt^[succ(i)] := l;
          j := son^[l];
          son^[l] := i;
          prnt^[j] := c;
          if (j < T) then prnt^[succ(j)] := c;
          son^[c] := j;
          c := l;
        end;
      c := prnt^[c];
    Until (c = 0);                     { Repeat it Until reaching the root }
  end;


Var
  code, len : Word;

  Procedure EncodeChar(c : Word; PutBytes : PutBytesProc);
  Var
    i : Word;
    j, k : Integer;
  begin
    i := 0;
    j := 0;
    k := prnt^[c + T];
    { search connections from leaf node to the root }
    Repeat
      i := i shr 1;
 {
        if node's address is odd, output 1
        else output 0
        }
      if Boolean(k and 1) then inc(i, $8000);
      inc(j);
      k := prnt^[k];
    Until (k = R);
    Putcode(j, i, PutBytes);
    code := i;
    len := j;
    update(c);
  end;




Procedure EncodePosition(C: Word;PutBytes: PutBytesProc);
Var
  I : Word;
  J : Word;

Begin

  { output upper 6 bits With encoding }
  I:=C ShR 006;
  J:=P_Code[I];
  Putcode(P_Len[I],J ShL 008,PutBytes);
  { output lower 6 bits directly }
  Putcode(006,(C And $3f) Shl 010,PutBytes);

End;




Procedure Encodeend(PutBytes: PutBytesProc);
Var
  Temp : Byte;
  Got  : Word;

Begin

  If Boolean(PutLen) Then
  Begin
    Temp:=Lo(PutBuf ShR 008);
    PutBytes(0,0,Temp,001,Got);
    Inc(CodeSize);
  End;

End;




Function DecodeChar(GetBytes : GetBytesProc) : Integer;
Var
  C : Word;

Begin

  C:=Son^[R];
  {  * start searching tree from the root to leaves.
     * choose node #(son[]) if input bit = 0
     * else choose #(son[]+1) (input bit = 1)}
  While (C<T) Do
  Begin
    C:=C+GetBit(GetBytes);
    C:=Son^[C];
  End;
  C:=C-T;
  Update(C);
  DecodeChar:=Integer(C);

End;




Function DecodePosition(GetBytes: GetBytesProc): Word;
Var
  C : Word;
  I : Word;
  J : Word;

Begin

  { decode upper 6 bits from given table }
  I:=GetByte(GetBytes);
  C:=Word(D_Code[I] ShL 6);
  J:=D_Len[I];
  { input lower 6 bits directly }
  Dec(J,2);
  While J<>000 Do
  Begin
    I:=(I ShL 001)+GetBit(GetBytes);
    Dec(J);
  End;
  DecodePosition:=C Or I And $3f;

End;




Procedure InitLZH;
Begin

  GetBuf:=000;
  GetLen:=000;
  PutLen:=000;
  PutBuf:=000;
  TextSize:=000;
  CodeSize:=000;
  PrintCount:=000;
  Match_Position:=000;
  Match_Length:=000;
  New(LSon);
  New(Dad);
  New(RSon);
  New(Text_Buf);
  New(Freq);
  New(Prnt);
  New(Son);

End;




Procedure EndLZH;
Begin

  Dispose(Son);
  Dispose(Prnt);
  Dispose(Freq);
  Dispose(Text_Buf);
  Dispose(RSon);
  Dispose(Dad);
  Dispose(LSon);

End;




Function  IsFileZipped(ZPath: ST79): Bool;
Var
  HandF : Hand;
  Stri0 : ST80;

Begin

  IsFileZipped:=BF;
  OFileF(HandF,ZPath,FRead,001);
  If IOResult2=000 Then
  Begin
    Stri0[000]:=Chr(007);
    BlockRead(HandF,Stri0[001],007,CntrHROW0);
    OFileF(HandF,'',FClose,001);
    If Stri0='KALPAK-' Then
    IsFileZipped:=BT;
  End;

End;




Procedure LZHPack(Var Bytes_Written : LongInt;
          GetBytes: GetBytesProc;PutBytes: PutBytesProc);
Var
  CT                : Byte;
  Got               : Word;
  I                 : IntM;
  Last_Match_Length : IntM;
  Len               : IntM;
  R                 : IntM;
  S                 : IntM;

Begin

  InitLZH;
  TextSize:=000;                     { rewind and rescan }
  StartHuff;
  InitTree;
  S:=000;
  R:=N-F;
  FillChar(Text_Buf^[000],R,' ');
  Len:=0;
  Got:=1;
  While (Len<F) And (Got<>0) Do
  Begin
    GetBytes(0,0,ct,1,Got);
    If Got<>0 Then
    Begin
      Text_Buf^[R+Len]:=CT;
      Inc(Len);
    End;
  End;
  TextSize:=Len;
  For I:=001 To F Do
  InsertNode(R-I);
  InsertNode(R);
  Repeat
    If (Match_Length>Len) Then
    Match_Length:=Len;
    If (Match_Length<=THRESHOLD) then
    Begin
      Match_Length:=001;
      EncodeChar(Text_Buf^[R],PutBytes);
    End
    Else
    Begin
      EncodeChar(255-THRESHOLD+Match_Length,PutBytes);
      EncodePosition(Match_Position,PutBytes);
    End;
    Last_Match_Length:=Match_Length;
    I:=000;
    Got:=001;
    While (I<Last_Match_Length) And (Got<>000) Do
    Begin
      GetBytes(0,0,CT,001,Got);
      If Got<>000 Then
      Begin
        DeleteNode(S);
        Text_Buf^[S]:=CT;
        If (S<Pred(F)) Then
        Text_Buf^[S+N]:=CT;
        S:=Succ(S) And Pred(N);
        R:=Succ(R) And Pred(N);
        InsertNode(R);
        Inc(I);
      End;
    End;
    Inc(TextSize,I);
    While (I<Last_Match_Length) Do
    Begin
      Inc(I);
      DeleteNode(S);
      S:=Succ(S) And Pred(N);
      R:=Succ(R) And Pred(N);
      Dec(Len);
      If Boolean(Len) Then InsertNode(R);
    End;
  Until (Len<=000);
  Encodeend(PutBytes);
  EndLZH;
  Bytes_Written:=TextSize;

End;




Procedure LZHUnpack(TextSize : LongInt;
          GetBytes: GetBytesProc;PutBytes: PutBytesProc);
Var
  A     : Byte;
  C     : IntM;
  C2    : Byte;
  Count : IntL;
  I     : IntM;
  J     : IntM;
  K     : IntM;
  Put   : Word;
  R     : IntM;

Begin

  InitLZH;
  InitTree;
  StartHuff;
  R:=N-F;
  FillChar(Text_Buf^[000],R,' ');
  Count:=000;
  While Count<TextSize Do
  Begin
    C:=DecodeChar(GetBytes);
    If (C<256) Then
    Begin
      C2:=Lo(C);
      PutBytes(Count,TextSize,C2,1,Put);
      Text_Buf^[R]:=C;
      Inc(R);
      R:=R And Pred(N);
      Inc(Count);
    End
    Else
    Begin
      I:=(R-Succ(DecodePosition(GetBytes))) And Pred(N);
      J:=C-255+THRESHOLD;
      For K:=000 To Pred(J) Do
      Begin
        C:=Text_Buf^[(I+K) And Pred(N)];
        C2:=Lo(C);
        PutBytes(Count,TextSize,C2,1,Put);
        Text_Buf^[R]:=C;
        Inc(R);
        R:=R And Pred(N);
        Inc(Count);
      End;
    End;
  End;
  EndLZH;

End;




Function  UnZip(SPath,TPath: ST79): Bool;
Var
  BErr  : Bool;
  BPass : Bool;
  Byte0 : Byte;
  Stri0 : ST10;
  Word0 : Word;
  Word1 : Word;

Procedure HelpUnScramble1;
Begin
  PGXLoad('SCRAMBLE','ONELINEK','','',000,000);
  If Not BooleanErr Then
  Begin
    FillField('SCRAMBLE',000,Scramble);
    MoveTXTToScreen('SCRAMBLE',002,001,
    'Image ist geschtzt durch ein Pawort. Bei der Eingabe des');
    MoveTXTToScreen('SCRAMBLE',002,005,
    ' Pawortes beachten Sie bitte die Gro-/Kleinschreibung.');
    PGXEdit('SCRAMBLE','ONELINEK','','',0009,0007,BT,000,BT,BT);
    If BooleanEx Then Exit;
    Scramble:=BuendigLR(EditorField);
  End;
End;

Procedure HelpUnScramble2;
Begin
  PGXLoad('SCRAMBLE','ONELINEK','','',000,000);
  If Not BooleanErr Then
  Begin
    FillField('SCRAMBLE',000,Scramble);
    MoveTXTToScreen('SCRAMBLE',002,001,
    'Drcken Sie bitte ENTER, um das Image zu laden, ...       ');
    MoveTXTToScreen('SCRAMBLE',002,005,
    ' oder ESC, um diesen Vorgang zu beenden.                ');
    PGXEdit('SCRAMBLE','ONELINEK','','',0009,0007,BT,000,BT,BT);
    If BooleanEx Then Exit;
    Scramble:=BuendigLR(EditorField);
  End;
End;

Begin

  TempI;
  BPass:=BT;
  UnZip:=BF;
  BVersion3:=BF;
  ZActivity:='Dekompression';
  Repeat
    BErr:=BF;
    If SPath=TPath Then
    Begin
      BErr:=BT;
      ZCopy(SPath,PRGPATH+'\7WCUNZIP.ZIP',BF);
      If Not BooleanErr Then SPath:=PRGPATH+'\7WCUNZIP.ZIP';
    End
    Else
    Begin
      OFileF(HIn,SPath,FRead,001);
      If IOResult2=000 Then
      Begin
        Stri0[000]:=Chr(008);
        BlockRead(HIn,Stri0[001],008,CntrHROW0);
        If Stri0='KALPAK-2' Then
        Begin
          If ZIPUNZIPC<000 Then
          HelpUnScramble2
          Else
          HelpUnScramble1;
        End
        Else
        If Stri0='KALPAK-3' Then
        Begin
          BVersion3:=BT;
          If ZIPUNZIPC<000 Then
          HelpUnScramble1
          Else
          Boolean13:=BT;
        End
        Else
        Boolean13:=BT;
        If Boolean13 Then
        If (Stri0='KALPAK-1')
        Or (Stri0='KALPAK-2')
        Or (Stri0='KALPAK-3') Then
        Begin
          OFileF(HOut,TPath,FNew,001);
          If IOResult2=000 Then
          Begin
            SHWB0:=BT;
            SegIn:=SWCSubPr.GetBuf(MaxBuf,Word0,Word1);
            MemErrorOut('UnZip',Word0,000);
            SegOut:=SWCSubPr.GetBuf(MaxBuf,Word0,Word1);
            MemErrorOut('UnZip',Word0,000);
            BlockRead(HIn,Size,004,CntrHROW0);
            If Stri0='KALPAK-3' Then
            Begin
              If ZIPUNZIPC<000 Then
              Begin
                ZIPUNZIPC:=256;
                LZHUnPack(Size,BlockGet,BlockPut);
              End
              Else
              Begin
                ZIPUNZIPC:=256;
                LZHUnPack(ZIPUNZIPC,BlockGet,BlockPut);
              End;
              If BooleanErr Then BPass:=BF;
            End
            Else
            If ZIPUNZIPC<000 Then
            LZHUnPack(Size,BlockGet,BlockPut)
            Else
            LZHUnPack(ZIPUNZIPC,BlockGet,BlockPut);
            BlockPut(Size,Size,Byte0,000,Temp);
            SWCSubPr.FreeBuf(SegIn,Word0);
            MemErrorOut('UnZip',Word0,000);
            SWCSubPr.FreeBuf(SegOut,Word0);
            MemErrorOut('UnZip',Word0,000);
            OFileF(HOut,'',FClose,001);
            PosNG:=000;
            PosNP:=000;
            UnZip:=BT;
            Buf:=000;
          End;
        End;
        OFileF(HIn,'',FClose,001);
        OFileF(HIn,PRGPATH+'\7WCUNZIP.ZIP',FErase,001);
      End;
    End;
  Until Not BErr;
  TempO;

End;



                                     {BoolMustZip}
Function  Zip(SPath,TPath: ST79;BAsk,BMZip: Bool): Bool;
Var
  BErr  : Bool;
  Byte0 : Byte;
  Stri0 : ST10;
  Stri1 : ST10;
  Word0 : Word;
  Word1 : Word;

Procedure HelpScramble;
Begin
  Scramble:='';
  ZActivity:='Kompression';
  If BAsk Then
  Begin
    Repeat
      PGXLoad('SCRAMBLE','ONELINEK','','',000,000);
      If Not BooleanErr Then
      Begin
        MoveTXTToScreen('SCRAMBLE',002,001,
        'Bitte geben Sie ein Pawort ein (Gro-/Kleinschreibung).');
        MoveTXTToScreen('SCRAMBLE',002,005,
        'Kein Pawort? Dann lassen Sie das Feld leer, dann ENTER.');
        PGXEdit('SCRAMBLE','ONELINEK','','',009,007,BT,000,BT,BT);
        If Not BMZip Then
        If BooleanEx Then Exit;
      End;
    Until Boolean13 Or (BMZip And BooleanEx);
    If Not BooleanEx Then
    Scramble:=BuendigLR(EditorField);
  End;
  Boolean13:=BT;
End;

Begin

  TempI;
  Zip:=BF;
  HelpScramble;
  BVersion3:=BF;
  If Boolean13 Then
  Begin
    Repeat
      BErr:=BF;
      If SPath=TPath Then
      Begin
        BErr:=BT;
        ZCopy(SPath,PRGPATH+'\7WCZIP.ZIP',BF);
        If Not BooleanErr Then SPath:=PRGPATH+'\7WCZIP.ZIP';
      End
      Else
      Begin
        OFileF(HIn,SPath,FRead,001);
        If IOResult2=000 Then
        Begin
          If Scramble<>'' Then
          Begin
            BVersion3:=BT;
            ZIPUNZIPC:=256;
            Stri0:='KALPAK-3';
          End
          Else
          Stri0:='KALPAK-1';
          Stri1[000]:=Chr(008);
          BlockRead(HIn,Stri1[001],008,CntrHROW0);
          If  (Stri1<>'KALPAK-1')
          And (Stri1<>'KALPAK-2')
          And (Stri1<>'KALPAK-3') Then
          Begin
            Seek(HIn,000);
            OFileF(HOut,TPath,FNew,001);
            If IOResult2=000 Then
            Begin
              SHWB0:=BT;
              SegIn:=SWCSubPr.GetBuf(MaxBuf,Word0,Word1);
              MemErrorOut('Zip',Word0,000);
              SegOut:=SWCSubPr.GetBuf(MaxBuf,Word0,Word1);
              MemErrorOut('Zip',Word0,000);
              BlockWrite(HOut,Stri0[001],008,CntrHROW0);
              Size:=Filesize(HIn);
              BlockWrite(HOut,Size,004,CntrHROW0);
              LZHPack(BWritten,BlockGet,BlockPut);
              BlockPut(BWritten,BWritten,Byte0,000,Temp);
              SWCSubPr.FreeBuf(SegIn,Word0);
              MemErrorOut('Zip',Word0,000);
              SWCSubPr.FreeBuf(SegOut,Word0);
              MemErrorOut('Zip',Word0,000);
              OFileF(HOut,'',FClose,001);
              Zip:=BT;
            End;
          End
          Else
          Zip:=BT;
          OFileF(HIn,'',FClose,001);
          OFileF(HIn,PRGPATH+'\7WCZIP.ZIP',FErase,001);
        End;
      End;
    Until Not BErr;
  End;
  TempO;

End;




END.
